#include "algorithm.hpp"
#include <iostream>
#include <fstream>
#include <sstream>
#include <stack>
#include <queue>

Graph ReadInGraph(std::string fileLocation) {
    std::cout << fileLocation << std::endl;
    std::ifstream graphFile(fileLocation);
    if (!graphFile.is_open()) {
        std::cerr << "Error opening file, check filename" << std::endl;
        exit(1);
    }
    Graph newGraph;
    graphFile >> newGraph.nodeCount;
    newGraph.vertices.resize(newGraph.nodeCount + 1);
    std::string line;
    int source, destination, weight;
    while (std::getline(graphFile, line)) {
        std::istringstream iss(line);
        iss >> source;
        do {
            iss >> destination >> weight;
            newGraph.vertices.at(source).nodeNumber = source;
            newGraph.AddEdge(source, destination, weight);
        } while (!iss.eof());
    }
    graphFile.close();
    return newGraph;
}

Graph::Graph(void) {
    nodeCount = -1;
    
}

Edge::Edge(Vertex* destination, int weight) {
    this->destination = destination;
    this->weight = weight;
} 

void Graph::AddEdge(int source, int destination, int weight) {
    vertices.at(source).edges.push_back(Edge(&(vertices.at(destination)), weight));
}

void Graph::PrintGraph(void) {
    for (int i = 1; i < vertices.size(); i++) {
        std::cout << "Source " << vertices.at(i).nodeNumber << " has ";
        for (Edge j : vertices.at(i).edges) { std::cout << " " << j.destination->nodeNumber; }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

void PrintSolution(std::vector<Vertex*> solution) {
    int cost = 0;
    std::cout << "Path: ";
    Vertex* lastVertex = solution.front();
    for (Vertex* i : solution) { 
        std::cout << i->nodeNumber << ' '; 
        for (Edge j : lastVertex->edges) {
            if (j.destination->nodeNumber == i->nodeNumber) {
                cost += j.weight;
            }
        }
        lastVertex = i;
    }
    std::cout << std::endl;
    std::cout << "Cost: " << cost << std::endl;
}

void UCBFS(Graph graph, int start) {
    std::queue<int> setds;
    std::vector<uint> dist(graph.nodeCount + 1, -1);
    setds.push(start);
    dist.at(start) = 0;
    while (!setds.empty())
    {
        int tmp = setds.front();
        setds.pop();
        for (Edge i : graph.vertices.at(tmp).edges)
        {
            int v = i.destination->nodeNumber;
            int weight = i.weight;
            if (dist[v] > dist.at(tmp) + weight)
            {
                if (dist[v] != ((uint)-1))
                    setds.pop();
                dist[v] = dist[tmp] + weight;
                setds.push(v);
            }
        }
    }
    std::cout << "Vertex\t Distance from Source" << std::endl;
    for (int i = 1; i < graph.nodeCount + 1; ++i) {
        std::cout << i << "\t\t" << dist[i] << std::endl;
    }
}

std::vector<Vertex*> BFS(Graph graph, int start, int end) {
    // Setup
    std::vector<Vertex*> solution;
    std::vector<bool> explored(graph.nodeCount + 1, 0);
    explored.at(start) = 1;
    std::queue<int> nodeQueue;
    nodeQueue.push(start);
    
    // Search
    while (!nodeQueue.empty()) {
        int currentNode = nodeQueue.front();
        nodeQueue.pop();
        solution.push_back(&(graph.vertices.at(currentNode)));
        if (currentNode == end) { break; }
        for (Edge i : graph.vertices.at(currentNode).edges) {
            if (!explored.at(i.destination->nodeNumber)) {
                explored.at(i.destination->nodeNumber) = 1;
                nodeQueue.push(i.destination->nodeNumber);
            }
        }
        if (solution.size() > graph.nodeCount + 1) { 
            std::cout << "Path too long" << std::endl;
            break;
        } 
    }
    return solution;
}

std::vector<Vertex*> DFS(Graph graph, int start, int end) {
    // Setup
    std::vector<Vertex*> solution;
    std::stack<int> nodeStack;
    std::vector<bool> explored(graph.nodeCount + 1, 0);
    nodeStack.push(start);

    // Search
    while (!nodeStack.empty()) {
        int currentNode = nodeStack.top();
        nodeStack.pop();
        solution.push_back(&(graph.vertices.at(currentNode)));
        if (currentNode == end) { break; }
        if (!explored.at(currentNode)) {
            explored.at(currentNode) = 1;
            for (Edge i : graph.vertices.at(currentNode).edges) {
                nodeStack.push(i.destination->nodeNumber);
            }
        }
    }
    return solution;
}
